import numpy as np
import crystally.core.constants as const


class FractionalVector:
    r"""A point in fractional space - e.g. in a crystal system.

    A fractional vector has three coordinates in range [0, 1). The fractional vector space is periodic, meaning
    that values greater or equal to one can be reduced: 1.1 == 0.1 and -0.2 == 0.8
    This class is only a wrapper for the numpy array which checks the coordinates and the shape of the array. To access
    the numpy array use :meth:`value`.

    :param vector: vector like container
    :return: FractionalVector with altered coordinates

    Examples
    --------
    >>> FractionalVector([1.0, 0.1, -0.6])
    FractionalVector([0.0, 0.1, 0.4])

    """

    def __init__(self, vector):
        vector = getattr(vector, "value", vector)
        self.__value = np.array(vector) % 1.0

    def __add__(self, other):
        other = getattr(other, "value", other)
        new_vec = FractionalVector(self.value + np.array(other))
        return new_vec

    def __sub__(self, other):
        return self + other*-1

    def __truediv__(self, other: float):
        new_vec = FractionalVector(self.value / other)
        return new_vec

    def __mul__(self, other):
        other = getattr(other, "value", other)
        new_vec = FractionalVector(self.value * np.array(other))
        return new_vec

    def __round__(self, n=None):
        if not n:
            n = const.VEC_ROUNDING
        return FractionalVector([n * round(x/n) for x in self])

    def __str__(self):
        return str(self.value)

    def __repr__(self):
        return "FractionalVector({})".format(self.value.tolist())

    def __getitem__(self, item):
        return self.value[item]

    def __len__(self):
        return len(self.value)

    def __hash__(self):
        return hash(tuple(round(self).tolist()))

    def __eq__(self, other):
        rounded1 = tuple(tuple(round(self).tolist()))
        rounded2 = tuple(tuple(round(other).tolist()))
        return rounded1 == rounded2

    @property
    def value(self):
        """ Get the wrapped numpy array

        :return: numpy array
        """
        return self.__value

    @value.setter
    def value(self, vec):
        raise TypeError("FractionalVector is immutable")

    @property
    def x(self):
        """ Get the x component of this vector, is identical to vec[0]

        :return: depends on the entry in the vector (probably a float)
        """
        if len(self) < 1:
            raise IndexError("This vector does not have a x component.")
        return self[0]

    @property
    def y(self):
        """ Get the y component of this vector, is identical to vec[0]

        :return: depends on the entry in the vector (probably a float)
        """
        if len(self) < 2:
            raise IndexError("This vector does not have a y component.")
        return self[1]

    @property
    def z(self):
        """ Get the z component of this vector, is identical to vec[0]

        :return: depends on the entry in the vector (probably a float)
        """
        if len(self) < 3:
            raise IndexError("This vector does not have a z component.")
        return self[2]

    def tolist(self):
        """ Unwrap vector and get list representation of numpy array

        :return: list of entries
        """
        return self.value.tolist()

    def diff(self, other):
        """ Get the minimal distance between two FractionalVector

        :param other: vector like container
        :return: 1xN numpy array where N is the dimension of the vectors

        Examples
        --------
        >>> FractionalVector([0.2, 0, 0]).diff([0, 0, 0.2])
        array([-0.2,  0. ,  0.2])

        >>> vec1 = FractionalVector([0.5, 0.5, 0.5])
        >>> vec2 = FractionalVector([0.6, 0.7, 0.8])
        >>> vec1.diff(vec2)
        array([0.1,  0.2 ,  0.3])
        >>> vec2.diff(vec1)
        array([-0.1,  -0.2 ,  -0.3])
        """
        other = FractionalVector(other)
        vec = np.array(other.value) - self.value
        return np.array([_min_periodic_distance(c) for c in vec])


def _min_periodic_distance(x):
    """Calculate the minimal periodic distance from the origin for one coordinate

    :param x: coordinate value
    :return: distance of coordinate from origin

    Examples
    --------
    >>>_min_periodic_distance(-0.6)
    0.4
    """
    if abs(x) > 0.5:
        return _min_periodic_distance(x - np.sign(x))
    else:
        return x




from crystally.core.vectors import *
import copy
import crystally.core.constants as constants
from typing import Union


class Atom:
    """An entry within a :class:`Lattice`.

    The Atom has three propertiers. An element from the periodic table, fractional coordinates in form of a
    :class:`~crystally.core.vectors.FractionalVector` and a sublattice name.

    :param element: string of element name
    :param position: fractional position of the atom
    :param sublattice: name of the sublattice the atom is in
    :return: Atom object
    """

    def __init__(self, element: str ="", position=(0, 0, 0), sublattice: str=""):
        self.element = str(element)
        self._position = FractionalVector(position)
        self.sublattice = str(sublattice)

    def __str__(self):
        string_representation = "Atom:"
        string_representation += f" {self.element:4s} "
        string_representation += f" [{', '.join([f'{coord:13.10f}' for coord in self.position.value])}] "
        string_representation += f" {self.sublattice:8s}"
        return string_representation

    def __repr__(self):
        rep = "Atom("
        rep += repr(self.element)    + ","
        rep += repr(self.position)   + ","
        rep += repr(self.sublattice) + ")"
        return rep

    def __eq__(self, other):
        return self.element == other.element        \
           and self.position == other.position      \
           and self.sublattice == other.sublattice

    @property
    def position(self):
        return self._position

    @position.setter
    def position(self, new_position):
        self._position = FractionalVector(new_position)


class Lattice:
    """A crystal lattice with atoms and crystal vectors

    :param atoms: string of element name
    :param vectors: fractional position of the atom
    :return: Lattice object
    """

    def __init__(self, vectors=np.identity(3), atoms=None):
        self._vectors = None
        self.atoms = list(atoms) if atoms is not None else list()
        self.vectors = vectors

    def __iter__(self):
        return iter(self.atoms)

    def __str__(self):
        string_rep = "Lattice:\n" \
                     + " "*5 + "Vectors:\n"
        for vec in self.vectors:
            components = ["{:13.10f}".format(i) for i in vec]
            string_rep += "{:>23s} {} {}\n".format(*tuple(components))
        string_rep += " "*5 + "Positions:\n"
        for i in range(0, len(self.atoms)):
            string_rep += " "*10
            string_rep += "{: <5d}".format(i)
            string_rep += str(self.atoms[i])
            string_rep += "\n"
        return string_rep

    def __repr__(self):
        rep = "Lattice("
        rep += repr(self.vectors.tolist())  + ","
        rep += repr(self.atoms)             + ")"
        return rep

    def __getitem__(self, item):
        return self.atoms[item]

    def __setitem__(self, key, value):
        self.atoms[key] = value

    def __delitem__(self, key):
        del(self.atoms[key])

    @property
    def vectors(self):
        return self._vectors

    @vectors.setter
    def vectors(self, vec):
        self._vectors = np.array(vec)

    @property
    def size(self):
        """ Calculates the volume of the lattice.

        :return: volume of lattice
        """
        if self.vectors.ndim == 1:
            return np.linalg.norm(self.vectors)
        elif self.vectors.ndim == 2:
            if len(self.vectors) == 1:
                return np.linalg.norm(self.vectors[0])
            elif len(self.vectors) == 2:
                return self.vectors[0].dot(self.vectors[1])
            elif len(self.vectors) == 3:
                return np.cross(self.vectors[0], self.vectors[1]).dot(self.vectors[2])
        else:
            raise NotImplemented("Size property is not compatible with lattice vectors!")

    def sort(self, key=None, tolerance=None):
        """ Sort the lattice with a given key.

        :param key: sorting argument which is used to sort the lattice. The following aruments can be used:

                    1. "element" - will sort according to the atoms element names
                    2. "position" - will sort according to the atoms absolute distance to the lattice origin
                    3. "sublattice" - will sort according to the atoms sublattice name
                    4. :class:`Lattice` - this will sort this lattice according to a different lattice. Equal atoms in
                       this and the other lattice should then have the same position in the atoms lists. Element and
                       Sublattice of the atoms are compared for equality. The position is compared with a tolerance in
                       angstrom. This means that if the atoms are closer together than the tolerance, their position
                       is treated as equal.

        :param tolerance: tolerance parameter which will be used when sorting this lattice with another lattice. If
                          None is specified the default value of constants.ABS_VEC_COMP_TOL is used.
        :return: nothing
        """
        if key is None:
            self.atoms.sort(key=lambda atom: (atom.element, atom.sublattice, self.distance(atom, [0.0, 0.0, 0.0])))
        elif key == "element":
            self.atoms.sort(key=lambda atom: atom.element)
        elif key == "position":
            self.atoms.sort(key=lambda atom: self.distance(atom, [0.0, 0.0, 0.0]))
        elif key == "sublattice":
            self.atoms.sort(key=lambda atom: atom.sublattice)
        elif isinstance(key, Lattice):
            self._sort_with_other_lattice(key, tolerance)
        else:
            self.atoms.sort(key=key)
        return None

    def find(self, element=None, sublattice=None, position=None, tolerance=None, first=False):
        """ Search the atom list according to the specified parameters.

        :param element: element name of the atoms to get
        :param sublattice: sublattice name of the atoms to get
        :param position: position of the atoms to get within a tolerance (can be specified)
        :param tolerance: tolerance of positional search in angstrom.
        :param first: If set to `True` the first found atom is returned.
        :return: list of :class:`Atoms <.Atom>` or :class:`Atoms <.Atom>` if first=True or `None` if no atoms satisfy
                 the criteria.
        """
        def element_cond(atom):
            return True if element is None else atom.element == element

        def sublattice_cond(atom):
            return True if sublattice is None else atom.sublattice == sublattice

        def position_cond(atom, tol=tolerance):
            tol = constants.ABS_VEC_COMP_TOL if not tol else tol
            return True if position is None else self.distance(atom, position) < tol

        found_atoms = (atom for atom in self.atoms
                       if element_cond(atom)
                       and sublattice_cond(atom)
                       and position_cond(atom))

        if first is True:
            try:
                return next(found_atoms)
            except StopIteration:
                return None
        else:
            return list(found_atoms)

    def remove(self, atom: Atom):
        """ Removes an atom from the atom list.

        :param atom: :class:`Atoms <.Atom>` that should be removed. This atom has to come from the atom list in the
                     first place.
        :return: None
        """
        for i, a in enumerate(self):
            if id(a) == id(atom):
                del(self[i])
                break
        else:
            raise ValueError("The specified atom is not in the atom list of this lattice.")

    def to_cartesian(self, position):
        """ Translates fractional coordinates to cartesian coordinates.

        :param position: fractional coordinates
        :return: cartesian coordinates
        """
        position = getattr(position, "position", position)
        return self.vectors.dot(np.array(position))

    def to_fractional(self, position, periodic=False):
        """ Translates cartesian coordinates to fractional coordinates.

        :param position: cartesian coordinates
        :param periodic: if periodic is set to `True`. The fractional coordinates are translated
                         to a value between 0 and 1.
        :return: fractional coordinates
        """
        position = getattr(position, "position", position)
        frac_position = np.linalg.inv(self.vectors) @ position
        if periodic:
            return FractionalVector(frac_position).value
        else:
            return frac_position

    def get_in_radius(self, center: Union[np.array, list, Atom], max_radius, min_radius=0.0):
        """ Get all :class:`.Atom` that are in range to a provided position.

        :param center: Search center in fractional coordinates. This can either be a list of coordinates or a :class:`.Atom`.
        :param max_radius: maximal radius in angstrom, that is searched
        :param min_radius: minimal radius in angstrom, that is searched. Default is 0 angstrom.

        :return: List of :class:`.Atom`
        """

        def condition(atom):
            return min_radius <= self.distance(atom, center) <= max_radius
        atom_list = [atom for atom in self.atoms if condition(atom)]
        atom_list.sort(key=lambda atom: self.distance(atom, center))
        return atom_list

    def index(self, position, tolerance=1e-3):
        """ Get the index within the atom list of the :class:`.Atom` at the provided position

        :param position: Position or Atom whose index is searched for
        :param tolerance: Tolerance in angstrom
        :return: index of :class:`.Atom`
        """
        position = getattr(position, "position", position)
        position = np.array(getattr(position, "value", position))
        for i in range(0, len(self.atoms)):
            if self.distance(self.atoms[i].position, position) < tolerance:
                return i
        return None

    def get_element_names(self):
        """ Get all occuring :attr:`Atom.element` names

        :return: List of strings
        """
        element_names = set()
        [element_names.add(x.element) for x in self.atoms]
        return list(element_names)

    def get_sublattice_names(self):
        """ Get all occuring :attr:`Atom.sublattice` names

        :return: List of strings
        """
        sublattice_names = set()
        [sublattice_names.add(x.sublattice) for x in self.atoms]
        return list(sublattice_names)

    def distance(self, position1, position2, periodic=True):
        """ Get the distance between two points

        :param position1: fractional coordinate as list or atom
        :param position2: fractional coordinate as list or atom
        :param periodic: bool, flag that indicates if the distance should be calculated under consideration of
                                periodic boundaries
        :return: float, distance in Angstrom
        """

        position1 = getattr(position1, "position", position1)
        position2 = getattr(position2, "position", position2)

        if periodic:
            position1 = FractionalVector(position1)
            position2 = FractionalVector(position2)
            diff_vector = position2.diff(position1)
        else:
            position1 = np.array(position1)
            position2 = np.array(position2)
            diff_vector = position2 - position1

        diff_vector_cartesian = diff_vector.dot(self.vectors)
        return np.sqrt(diff_vector_cartesian.dot(diff_vector_cartesian))

    def angle(self, position1, position2, center=(0, 0, 0), periodic=True):
        """ Get the angle between two positions and a center.

        :param position1: fractional coordinate as list or atom
        :param position2: fractional coordinate as list or atom
        :param center: fractional coordinate as list or atom
        :param periodic: bool, flag that indicates if the distance should be calculated under consideration of
                                periodic boundaries
        :return: float, angle in rad
        """

        position1 = getattr(position1, "position", position1)
        position2 = getattr(position2, "position", position2)
        center = getattr(center, "position", center)

        if periodic:
            position1 = FractionalVector(position1)
            position2 = FractionalVector(position2)
            center = FractionalVector(center)
            vec1 = center.diff(position1)
            vec2 = center.diff(position2)
        else:
            position1 = np.array(position1)
            position2 = np.array(position2)
            center = np.array(center)
            vec1 = position1 - center
            vec2 = position2 - center

        vec_cart1 = vec1.dot(self.vectors)
        vec_cart2 = vec2.dot(self.vectors)
        magn1 = np.sqrt(vec_cart1.dot(vec_cart1))
        magn2 = np.sqrt(vec_cart2.dot(vec_cart2))
        return np.arccos(vec_cart1.dot(vec_cart2)/(magn1*magn2))

    def increase_distance_rel(self, center, position, rel_increase):
        """ Function to easily manipulate lattices. A center and second position need to be provided. The second
        position is then shifted away from the center position by a provided percentage. The new position is then returned.
        Mind that no variables are changed by this function.

        :param center: fractional coordinates of the center position. Can be a list of coordinates or a :class:`.Atom`.
        :param position: fractional coordinates of the second position. Can be a list of coordinates or a :class:`.Atom`.
        :param rel_increase: distance increase in percent
        :return: new position as :class:`~crystally.core.vectors.FractionalVector`
        """
        # First check if atoms were passed to the function
        center = getattr(center, "position", center)
        position = getattr(position, "position", position)

        # Now convert the vectors to ensure periodicity
        center = FractionalVector(center)
        position = FractionalVector(position)

        # calculate the distance in fractional coordinates from center to position
        diff_vector = center.diff(position)

        # convert everything to cartesian coordinates
        center = self.vectors.dot(center.value)
        diff_vector = self.vectors.dot(diff_vector)

        # calculate the new position in cartesian coordinates
        new_position = center + diff_vector * (1+rel_increase)

        # convert the new position to fractional coordinates
        return FractionalVector(np.linalg.inv(self.vectors).dot(new_position))

    def increase_distance_abs(self, center, position, abs_increase):
        """ Function to easily manipulate lattices. A center and second position need to be provided. The second
        position is then shifted away from the center position by a provided value. The new position is then returned.
        Mind that no variables are changed by this function.

        :param center: fractional coordinates of the center position. Can be a list of coordinates or a :class:`.Atom`.
        :param position: fractional coordinates of the second position. Can be a list of coordinates or a :class:`.Atom`.
        :param abs_increase: Shift in angstrom
        :return: new position as :class:`~crystally.core.vectors.FractionalVector`
        """
        distance = self.distance(center, position)
        rel_distance = abs_increase/distance
        return self.increase_distance_rel(center, position, rel_distance)

    def diff(self, other, tolerance=None):
        """ Find the atomic differences bewteen this and another lattice. Atoms are compared according to element
            sublattice and position. Mind that the fractional position of this and the other lattice are converted
            to cartesian coordinates with the vectors of this lattice.

        :param other: other :class:`.Lattice`.
        :param tolerance: tolerance in angstrom with which the atoms position are compared.
        :return: tuple with two lists, the first list consists of the atoms that were in this lattice, but were not
                 found in the other lattice. The second list contains the atoms that were in the other lattice, but not
                 found in this lattice.
        """

        if not tolerance:
            tolerance = const.ABS_VEC_COMP_TOL

        self_not_found = []
        other_not_found = list(other.atoms)

        for atom1 in self:
            for atom2_id, atom2 in enumerate(other_not_found):
                if self._compare_atoms(atom1, atom2, tolerance):
                        del(other_not_found[atom2_id])
                        break
            else:
                self_not_found.append(atom1)

        return self_not_found, other_not_found

    def _sort_with_other_lattice(self, other, tolerance=None):
        if not tolerance:
            tolerance = const.ABS_VEC_COMP_TOL

        old_order = list(self.atoms)
        new_order = []
        for atom1 in other:
            for atom2_id, atom2 in enumerate(old_order):
                if self._compare_atoms(atom1, atom2, tolerance):
                    new_order.append(atom2)
                    del(old_order[atom2_id])
        new_order += old_order
        self.atoms = new_order

    def _compare_atoms(self, atom1, atom2, dist_tolerance):
        if atom1.element != atom2.element:
            return False
        if atom1.sublattice != atom2.sublattice:
            return False
        if self.distance(atom1, atom2) > dist_tolerance:
            return False
        return True

    @staticmethod
    def from_lattice(lattice, size_x: int = 1, size_y: int = 1, size_z: int = 1):
        """ Create a new lattice with an already existing one. You can also provide size values to expand the lattice.

        :param lattice: the :class:`.Lattice` on which the new lattice is based on
        :param size_x: int, multiplier by which the base lattice is expanded in x direction
        :param size_y: int, see size_x
        :param size_z: int, see size_x
        :return: new :class:`.Lattice`
        """

        new_lattice_size = np.array([size_x, size_y, size_z])
        new_lattice_positions = []

        def reorientate_position(pos, x, y, z): return FractionalVector((pos + np.array([x, y, z])) / new_lattice_size)

        lattice_expansion = ((x, y, z) for x in range(size_x) for y in range(size_y) for z in range(size_z))
        new_lattice_positions += [Atom(element=atom.element,
                                       sublattice=atom.sublattice,
                                       position=reorientate_position(atom.position.value, *coord))
                                  for coord in lattice_expansion for atom in lattice.atoms]

        lattice_vectors = lattice.vectors * new_lattice_size
        new_lattice = Lattice(lattice_vectors, new_lattice_positions)
        new_lattice.sort("element")
        return new_lattice

    @staticmethod
    def grid_atoms(lattice, grid_step):

        def to_grid(lattice, position, round_func):
            position = lattice.vectors @ position
            return tuple(int(x) for x in round_func(np.round(position / grid_step, 10)))

        grid_size = to_grid(lattice, np.array([1, 1, 1]), np.ceil)

        grid = [[[[] for _ in range(grid_size[2])]
                     for _ in range(grid_size[1])]
                     for _ in range(grid_size[0])]

        for atom in lattice:
            x, y, z = to_grid(lattice, atom.position, np.floor)
            grid[x][y][z].append(atom)
        return grid


def concat_lattices(lattice1: Lattice, lattice2: Lattice, direction: int):
    for i in range(lattice1.vectors.shape[0]):
        if i == direction:
            continue
        if not np.allclose(lattice1.vectors[i], lattice2.vectors[i], atol=1e-10):
            raise ValueError("shape of lattices does not match: "
                             "lattice1: {} lattice2: {}".format(str(lattice1.vectors[i]), lattice2.vectors[i]))
    new_lattice = Lattice(vectors=lattice1.vectors, atoms=[])
    new_lattice.vectors[direction] = lattice1.vectors[direction] + lattice2.vectors[direction]
    for atom in lattice1.atoms:
        new_atom = copy.copy(atom)
        new_atom.position = FractionalVector(atom.position.value.dot(lattice1.vectors).dot(np.linalg.inv(new_lattice.vectors)))
        new_lattice.atoms.append(new_atom)

    for atom in lattice2.atoms:
        new_atom = copy.copy(atom)
        shift = new_atom.position.value * 0
        shift[direction] = 1
        shift_cart = shift.dot(lattice1.vectors)
        new_position = new_atom.position.value.dot(lattice2.vectors) + shift_cart
        new_position = new_position.dot(np.linalg.inv(new_lattice.vectors))
        new_atom.position = FractionalVector(new_position)
        new_lattice.atoms.append(new_atom)

    return new_lattice





from crystally.core.lattice import *


def make_neb(lattice1: Lattice, lattice2: Lattice, image_numbers=1, tolerance=None):
    """ Linearly interpolate the atom list two lattices and new lattices according to the number of images.

    :param lattice1: first :class:`~crystally.core.lattice.Lattice`
    :param lattice2: second :class:`~crystally.core.lattice.Lattice`
    :param image_numbers: number of images for the NEB algorithm
    :param tolerance: the tolerance in angstrom which is used to compare atoms and sort the lattices
    :return: list of new :class:`~crystally.core.lattice.Lattice`. This includes the lattices for the initial and
    final state.
    """

    if len(lattice1.atoms) != len(lattice2.atoms):
        raise ValueError("lattices must have the same number of atoms!")

    # make copies of original lattices
    lattice1 = Lattice.from_lattice(lattice1)
    lattice2 = Lattice.from_lattice(lattice2)

    # sort lattices to each other
    lattice1.sort(lattice2, tolerance=tolerance)
    lattice2.sort(lattice1, tolerance=tolerance)

    neb_lattices = [lattice1]

    # the weighting for the interpolation is determined by the number of images
    for weighting in np.linspace(0, 1, image_numbers+2)[1:image_numbers+1]:
        new_lattice = Lattice(vectors=lattice1.vectors, atoms=[])
        for atom1, atom2 in zip(lattice1, lattice2):
            new_position = atom1.position + atom1.position.diff(atom2.position) * weighting
            new_atom = Atom(element=atom1.element, position=new_position, sublattice=atom1.sublattice)
            new_lattice.atoms.append(new_atom)
        neb_lattices.append(new_lattice)

    neb_lattices.append(lattice2)
    return neb_lattices


def get_element_number_list(lattice):
    """Get a two dimensional list with the sort order of the elements within the lattice.

    The first column specifies the element name and the second column the number of adjacent atoms with this
    element. This information can be used for VASP Input files.
    For instance if the sort order of the atoms is as follows:

    X, X, Y, Y, Y, X, X

    (where X and Y are the elements of the atoms)
    the function would return the following table:

    =============  =============================
    Element Name   Number of repeating elements
    =============  =============================
    X              2
    Y              4
    X              2
    =============  =============================

    :return: two dimensional list - the first column is an int, the second a string

    Examples
    --------
    >>> #lattice = generate_from_crystal(ceria(),2,2,2).sort("position")
    >>> #print(lattice.get_element_number_list())
    [['Ce', 1], ['O', 1], ['Ce', 3], ['O', 7]]
    """
    element_number_list = []
    for atom in lattice.atoms:
        if not element_number_list:
            element_number_list.append([atom.element, 1])
        elif element_number_list[-1][0] == atom.element:
            element_number_list[-1][1] += 1
        else:
            element_number_list.append([atom.element, 1])
    return element_number_list
